/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | Multiphase Code Repository by HZDR
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is part of the Multiphase Code Repository by HZDR.

    Copyright (C) 2024 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
    Website: https://hzdr.de

    Multiphase Code Repository by HZDR is based on the free software for
    computational fluid dynamics (CFD) from the OpenFOAM Foundation.
    Copyright (C) 2024 by OpenFOAM Foundation, Website: https://openfoam.org

    If you are interested in which files are original OpenFOAM Foundation
    files, which OpenFOAM Foundation files were modified, and which files were
    newly created, see FILES.md.

    Multiphase Code Repository by HZDR is free software: you can redistribute
    it and/or modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Multiphase Code Repository by HZDR is distributed in the hope that it will
    be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
    Public License for more details.

    You should have received a copy of the GNU General Public License along
    with the Multiphase Code Repository by HZDR. If not, see
    <http://www.gnu.org/licenses/>.

Class
    Foam::functionObjects::populationBalanceSetSizeDistribution

Description
    Sets the population balance size distribution by overwriting the values in
    the size-group fraction fields with values obtained by integrating a given
    distribution.

    This function will only execute in post-processing mode.

    Example of function object specification:
    \verbatim
    populationBalanceSetSizeDistribution
    {
        type        populationBalanceSetSizeDistribution;
        libs        ("libaddonmultiphaseEulerFunctionObjects.so");

        populationBalance bubbles;

        //phase       air1; // <-- alternative to specifying
                            //     'populationBalance' if only a single phase's
                            //     size group fractions are to be set

        distribution
        {
            type        tabulatedDensity;
            Q           3;
            file        "constant/initialDistribution.bubbles";
        }
    }
    \endverbatim

Usage
    \table
        Property          | Description             | Required        | Default
        populationBalance | population balance name | if phase is not \
                                                      specified       |
        phase             | phase name              | if populationBalance is \
                                                      not specified   |
        distribution      | the distribution        | yes             |
    \endtable

See also
    Foam::diameterModels::populationBalanceModel
    Foam::functionObjects::fvMeshFunctionObject
    Foam::functionObject

SourceFiles
    populationBalanceSetSizeDistribution.C

\*---------------------------------------------------------------------------*/

#ifndef populationBalanceSetSizeDistribution_H
#define populationBalanceSetSizeDistribution_H

#include "fvMeshFunctionObject.H"
#include "populationBalanceModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace functionObjects
{

/*---------------------------------------------------------------------------*\
             Class populationBalanceSetSizeDistribution Declaration
\*---------------------------------------------------------------------------*/

class populationBalanceSetSizeDistribution
:
    public fvMeshFunctionObject
{
private:

    // Private Data

        //- Name of the population balance
        const word popBalName_;

        //- Name of the phase
        const word phaseName_;

        //- Distribution
        autoPtr<distribution> distribution_;


public:

    //- Runtime type information
    TypeName("populationBalanceSetSizeDistribution");


    // Constructors

        //- Construct from Time and dictionary
        populationBalanceSetSizeDistribution
        (
            const word& name,
            const Time& runTime,
            const dictionary& dict
        );

        //- Disallow default bitwise copy construction
        populationBalanceSetSizeDistribution
        (
            const populationBalanceSetSizeDistribution&
        ) = delete;


    //- Destructor
    virtual ~populationBalanceSetSizeDistribution();


    // Member Functions

        //- Return the list of fields required
        virtual wordList fields() const
        {
            return wordList::null();
        }

        //- Execute. Does nothing.
        virtual bool execute();

        //- Calculate and write the size group fraction fields
        virtual bool write();


    // Member Operators

        //- Disallow default bitwise assignment
        void operator=(const populationBalanceSetSizeDistribution&) = delete;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace functionObjects
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
