/*---------------------------------------------------------------------------*\
  ==  == ====== ====   ====    |
                    \\     ||  | Multiphase Code Repository by HZDR
  ======   //   ||  || ===//   | Website: https://doi.org/10.14278/rodare.767
  ||  ||  //    ||  // || \\   | License: GPL-3.0-or-later
  ==  == ====== ====   ==  ==  |
-------------------------------------------------------------------------------
License
    This file is part of the Multiphase Code Repository by HZDR.

    Copyright (C) 2024 by Helmholtz-Zentrum Dresden-Rossendorf e.V. (HZDR),
    Website: https://hzdr.de

    Multiphase Code Repository by HZDR is based on the free software for
    computational fluid dynamics (CFD) from the OpenFOAM Foundation.
    Copyright (C) 2024 by OpenFOAM Foundation, Website: https://openfoam.org

    If you are interested in which files are original OpenFOAM Foundation
    files, which OpenFOAM Foundation files were modified, and which files were
    newly created, see FILES.md.

    Multiphase Code Repository by HZDR is free software: you can redistribute
    it and/or modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Multiphase Code Repository by HZDR is distributed in the hope that it will
    be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
    Public License for more details.

    You should have received a copy of the GNU General Public License along
    with the Multiphase Code Repository by HZDR. If not, see
    <http://www.gnu.org/licenses/>.

Class
    Foam::phaseTransferModels::attachmentModel

Description
    Abstract base class for attachment models for phase transfer from a
    multicomponent detached phase (might have only single specie) to the
    corresponding specie in a multicomponent aggregate phase.

SourceFiles
    attachmentModel.C

\*---------------------------------------------------------------------------*/

#ifndef attachmentModel_H
#define attachmentModel_H

#include "phaseTransferModel.H"
#include "rhoFluidMulticomponentThermo.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

namespace phaseTransferModels
{

/*---------------------------------------------------------------------------*\
                     Class attachmentModel Declaration
\*---------------------------------------------------------------------------*/

class attachmentModel
:
    public phaseTransferModel
{
private:

    // Private Member Functions

        volScalarField& attachmentRatio
        (
            const volScalarField& dB,
            const volScalarField& dP,
            const volScalarField& Yc,
            const volScalarField& Yi,
            const volScalarField& rhoC,
            const volScalarField& rhoI
        ) const;


protected:

    // Protected Data

        //- Interface
        const dispersedPhaseInterface interface_;

        //- Direction of phase transfer
        const scalar sign_;

        //- Const reference to the detached phase
        const phaseModel& detachedPhase_;

        //- Const reference to the aggregate phase
        const phaseModel& aggregatePhase_;

        //- Residual value of alpha of the aggregate phase below which no mass
        //  transfer takes place
        const dimensionedScalar residualAlpha_;

        //- Aggregate phase thermo model
        const rhoFluidMulticomponentThermo& thermo_;

        //- List of species to return for phaseTransferModel
        const hashedWordList species_;

        //- The name of the carrier species
        const word carrierSpecies_;

        //- Index of carrier specie in aggregate phase
        const label indexC_;

        //- Index of first specie in aggregate phase
        const label indexI_;

        //- Maximum allowed bubble surface loading
        const scalar maxLoading_;

        //- Number of attached particles per bubble/droplet
        mutable volScalarField attachmentRatio_;

        //- Surface loading
        mutable volScalarField surfaceLoading_;


    // Protected Member Functions

        volScalarField& surfaceLoading
        (
            const volScalarField& dB,
            const volScalarField& dP,
            const volScalarField& Yc,
            const volScalarField& Yi,
            const volScalarField& rhoC,
            const volScalarField& rhoI
        ) const;

        //- Limiter for attachment rate
        virtual tmp<volScalarField> attachmentLimiter() const;

        //- Attachment rate
        virtual tmp<volScalarField> rate() const = 0;


public:

    // Constructors

        //- Construct from components
        attachmentModel
        (
            const dictionary& dict,
            const phaseInterface& interface
        );


    //- Destructor
    virtual ~attachmentModel();


    // Member Functions

        //- The list of individual species that are transferred
        virtual const hashedWordList& species() const;

        //- The mass transfer rate for individual species
        virtual HashPtrTable<volScalarField> dmidtf() const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace phaseTransferModels
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
